package org.zhouw.stps;

import java.util.Hashtable;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.locks.*;
import org.zhouw.stps.tp.*;
import org.zhouw.stps.msg.MsgUtil;
import org.zhouw.stps.msg.SubTransactionGroupProtos.*;

/**
 * Used for transaction executor to dispatch sub-transactions to TPS nodes
 * Multiple sub-transactions belongs to same node should be grouped together
 * and submitted in one message and execute as one local transaction
 * 
 * To prevent deadlock, in STG class, it would not acquire another lock except itself
 * @author zhouw
 *
 */
public class SubTransactionGroup {

	public void terminateForMM(){
		if(this.trans!=null){
			for(SubTransaction t:this.trans){
				if(t.rootEntry!=null){
					t.rootEntry.state.transactionLeaves();
				} else {
					;//GlobalPara.logger.warning("[Error]RootEntry not exist");
				}
			}
		}
	}
	
	public STGMsg getMsgBody()
	{
		STGMsg.Builder msgBuilder=STGMsg.newBuilder();
		msgBuilder.setGID(this.GID)
		.setGTPNodeID(this.GTPNodeID)
		.setId(this.id)
		.setIsReadOnly(this.isReadOnly)
		.setKey(this.key)
		.setTimestamp(this.timestamp)
		.setTPnodeID(this.TPnodeID);
		
		for(int i=0;i<this.trans.length;i++)
			msgBuilder.addTrans(this.trans[i].getMsgBody());
		
		return msgBuilder.build();
	}
	
	public SubTransactionGroup()
	{
		this.state=STGState.START;
	}
	
	public Vector<SubTransaction> getDerived()
	{
		if(this.trans==null)
			return null;
		Vector<SubTransaction> result=new Vector<SubTransaction>();
		for(SubTransaction st: this.trans){
			if(st.derivedTransactions!=null)
				result.addAll(st.derivedTransactions);
		}
		
		return result;
	}
	
	//generate key to be unique at the LTM
	public String genKey()
	{
		this.key=this.GID+"-" +this.id;
		return this.key;
	}
	
	public SubTransactionGroup(STGMsg msg)
	{
		this.timestamp=msg.getTimestamp();
		this.GID=msg.getGID();
		this.isReadOnly=msg.getIsReadOnly();
		this.id=msg.getId();
		this.TPnodeID=msg.getTPnodeID();
		this.GTPNodeID=msg.getGTPNodeID();
		this.key=msg.getKey();
		this.isCommitOK=new boolean[msg.getTransCount()];
		this.isCommitCompleted=new boolean[msg.getTransCount()];
		this.isCheckpointCompleted=new boolean[msg.getTransCount()];
		this.state=STGState.START;
		
		List<STMsg> list=msg.getTransList();
		this.trans=new SubTransaction[list.size()];
		for(int i=0;i<list.size();i++) {
			STMsg st=list.get(i);
			String className=st.getClassName();
			String rowKey=st.getStrRowKey();
			Hashtable<String, String> parameters=null;
			if(st.getParasCount()>0)
			{
				parameters=new Hashtable<String, String>(st.getParasCount()*2);
				for(int j=0;j<st.getParasCount();j++)
					parameters.put(st.getParas(j).getKey() , st.getParas(j).getValue());
			}
			SubTransaction t=SubTransaction.buildSubTransactionByName(className, st.getStrDBName(), st.getStrTblName(), rowKey, parameters);
			if(t==null) {
				GlobalPara.logger.severe("Null ST is constructed when Constructing STG from messages");
			}else{
				this.trans[i]=t;
				this.trans[i].parentGroup=this;
			}
		}
	}
	
	/**
	 * it has to be locked when any subtransaction want to execute
	 */
	//public ReentrantLock lock=new ReentrantLock() ;
	
	public enum STGMSGTYPE{DOSTART,REPORTCOMMIT,REPORTABORT,REPORTRESTART,
		DORESTART, DOABORT, DATACOMMITDONE}  //DOCOMMIT
	public STGState state;
	private boolean[] isCommitOK;
	private boolean[] isCommitCompleted;
	private boolean[] isCheckpointCompleted;
	
	public long timestamp;
	public boolean isReadOnly;
	
	/**
	 * The ID the belonging global transaction
	 */
	public long GID;
	
	/**
	 * index of this STG within the all STGs of the global transaction
	 */
	public int id;
	
	/**
	 * Index as it is inserted into the pool of LTM
	 */
	public String key;
	/**
	 * the node where this STG is executed
	 */
	public int TPnodeID;
	/**
	 * global Transaction coordinator
	 */
	public int GTPNodeID;
	
	public long getTimeStamp()
	{
		return this.timestamp;
	}
	
	public void setTimeStamp(long newTimestamp)
	{
		this.timestamp=newTimestamp;
	}

	/**
	 * @param timestamp
	 * @param id
	 * @return
	 */
	public String getSTGkey()
	{
		return key;
	}
	
	public SubTransactionGroup(long GID, long timestamp, SubTransaction[] t, int id, int TPnodeID, int GTPNodeID)
	{
		this.GID=GID;
		this.timestamp=timestamp;
		this.trans=t;
		for(int i=0;i<t.length;i++)
			this.trans[i].parentGroup=this;
		//this.order=this.getSequence();
		this.id=id;
		this.TPnodeID=TPnodeID;
		this.GTPNodeID=GTPNodeID;
		this.isCommitOK=new boolean[this.trans.length];
		this.isCommitCompleted=new boolean[this.trans.length];
		this.isCheckpointCompleted=new boolean[this.trans.length];
		
		this.state=STGState.START;
		this.genKey();
	}
	
	public SubTransaction[] trans;
	//order to process the rows
	//public final int[] order;
	

	
	private String printStatusDiff(STGState current, STGMSGTYPE msgType)
	{
		return this.key + ":"+ this.timestamp+ "| Receive '"+ msgType.toString() +"' Message in '"+current.toString() +"' Status"; 
	}
	
	private boolean isAnyDataToCommit()
	{
		boolean isDo=false;
		for(int i=0;i<this.trans.length;i++) {
			if(this.trans[i].DataToPut!=null||this.trans[i].DataToDelete!=null)
			{
				isDo=true;
				break;
			}
		}
		return isDo;
	}
	
	/**
	 * No need to lock as it is only invoked when all STs are ready to commit
	 * @return
	 */
	public ResultSetMsg constructCommitResult()
	{
		ResultSetMsg.Builder msgbuilder=ResultSetMsg.newBuilder();
		
		for(int n=0;n<this.trans.length;n++) {
			SubTransaction t = this.trans[n];
			String[][] s=t.DataToReturn;
			if(s!=null) {
				for(int i=0;i<s.length;i++) {
					Parameter.Builder p=Parameter.newBuilder();
					p.setKey(s[i][0]);
					p.setValue(s[i][1]);
					msgbuilder.addResult(p.build());
				}
			}
			UpdateOnDataRowMsg.Builder updateBuilder=UpdateOnDataRowMsg.newBuilder();
			updateBuilder.setStrRowKey(t.strRowKey).setStrTblName(t.strTblName)
			.setNodeID(t.parentGroup.TPnodeID).setStgID(t.parentGroup.id);

			s=t.DataToPut;
			if(s!=null) {
				for(int i=0;i<s.length;i++) {
					Parameter.Builder p=Parameter.newBuilder();
					p.setKey(s[i][0]);
					p.setValue(s[i][1]);
					updateBuilder.addDataToPut(p.build());
				}
			}
			
			if(t.rootEntry.latestWriteTransaction==-1) {//the data item is loaded the first time, whole data row is loaded
				if(!t.isUpdateSameDataItem){
					updateBuilder.setIsContainsDataRow(true);
					s=t.rootEntry.committedData.getRowData();
					if(s!=null&&s.length>0) {
						for(int i=0;i<s.length;i++) {
							Parameter.Builder p=Parameter.newBuilder();
							p.setKey(s[i][0]);
							p.setValue(s[i][1]);
							updateBuilder.addDataRow(p.build());
						}
					}
				}
			}

			if(t.DataToDelete!=null) {
				for(int i=0;i<t.DataToDelete.length;i++)
					updateBuilder.addDataToDelete(t.DataToDelete[i]);
			}

			updateBuilder.setTimestamp(t.parentGroup.timestamp);
			msgbuilder.addUpdates(updateBuilder.build());
		}
		ResultSetMsg msg=msgbuilder.build();
		return msg;
	}
	
	/**
	 * Initialize the STG data structure, creating the data items, adding subtransactions
	 * load the data item if not loaded 
	 * @return 1: success, 0, restart, -1: error
	 */
	public int doStart()
	{
		if(this.state!=STGState.START){
			GlobalPara.logger.warning("The STG can not perform dostart in the status: "+ this.state);
			return -1;
		}
			
		for(int i=0;i<this.isCommitOK.length;i++){
			this.isCommitOK[i]=false;
			this.isCommitCompleted[i]=false;
			this.isCheckpointCompleted[i]=false;
		}
		
		this.state=STGState.WAIT;
		return LTMActions.AddGroupSTs(this.trans, false);
	}
	
//	public int doneCommit(long timestamp)
//	{
//		try{
//			return this.handleMsg(timestamp, STGMSGTYPE.DATACOMMITDONE, -1, null);
//		} finally{ lock.unlock(); }
//	}
//	
//	/**
//	 * 
//	 * @param timestamp
//	 * @return 0 message ignored, 1 it enters committing state, 2 it is committed 
//	 */
//	public int doCommit(long timestamp)
//	{
//		lock.lock();
//		try{
//			return this.handleMsg(timestamp, STGMSGTYPE.DOCOMMIT, -1, null);
//		} finally{
//			lock.unlock();
//		}
//	}
	
	/**
	 * @return 1: success, 0, still need to restart, -1 error
	 */
	public int doRestart(long newTimestamp)
	{
		int restartResult=1;
		if(newTimestamp<=this.timestamp) {
			GlobalPara.logger.warning("[Wrong Message Content]New TimeStamp is smaller than current, STG "+this.timestamp+" refuese to restart");
			return -1;
		}
		if(this.state!=STGState.ASKRESTART&&this.state!=STGState.ASKCOMMIT&&this.state!=STGState.WAIT){
			GlobalPara.logger.warning("[Error]STG "+this.key+" is not in the right status to restart, its state is: " + this.state);
			return -1;
		} else {
			this.timestamp=newTimestamp;
			if(this.state!=STGState.ASKRESTART){
				for(int i=0;i<this.trans.length;i++) {
					this.trans[i].rootEntry.removeSubTransaction(this.trans[i], true);
//					if(this.isCommitOK[i]){
//						this.isCommitOK[i]=false;
//						if(this.trans[i].derivedTransactions!=null)
//							this.trans[i].derivedTransactions.clear();
//						this.trans[i].rootEntry.committedData.restoreToBackup();
//					}
				}	
			}
			
			for(SubTransaction st:this.trans){
				st.isUpdateSameDataItem=false;
			}
			this.state=STGState.WAIT;
			restartResult=LTMActions.AddGroupSTs(this.trans, true);
			if(restartResult!=1){
				this.state=STGState.ASKRESTART;
			}
		}
		return restartResult;
	}
	
	/**
	 * This operation will work in any timestamps, even when it is different
	 * @param timestamp
	 * @return
	 */
	public int doAbort()
	{
		if(this.state==STGState.COMMITTED||this.state==STGState.ABORT||this.state==STGState.START){
			GlobalPara.logger.warning("The STG can not perform abort in the status: "+ this.state);
			return -1;
		}
			
		for(SubTransaction st:this.trans){
			LTMActions.RemoveOneST(st, false);
		}
		
		this.state=STGState.ABORT;
		
		return 0;
	}
	
	public int doCommit()
	{
		if(this.state!=STGState.ASKCOMMIT){
			GlobalPara.logger.warning("The STG can not perform docommit action in the status: "+ this.state);
			return -1;
		}
			
		this.state=STGState.COMMITTING;
		for(SubTransaction st:this.trans){
			LTMActions.CommitOneST(st);
		}
		
		return 0;
	}
	
	public int reportAbort(SubTransaction reportST)
	{
		if(this.state!=STGState.WAIT){
			GlobalPara.logger.warning("The STG can not perform reportAbort action in the status: "+ this.state);
			return -1;
		}
		MsgBroker.reportAbort(this);
		
		return 0;
	}
	public int reportCommit(long timestamp, SubTransaction t)
	{
		if(this.state!=STGState.WAIT){
			GlobalPara.logger.warning("The STG can not perform reportcommit action in the status: "+ this.state);
			return -1;
		}
		for(int i=0;i<trans.length;i++)
			if(t==trans[i]) {
				isCommitOK[i]=true;
				break;
			}
		if(this.isTotalCommitable()) {
			MsgBroker.reportCommit(this);
			this.state=STGState.ASKCOMMIT;
		}
		return 0;
	}
	
	
	public void reportCommitDone(SubTransaction t)
	{
		for(int i=0;i<trans.length;i++)
			if(t==trans[i]) {
				this.isCommitCompleted[i]=true;
				break;
			}
		checkIfTerminatable();
	}
	
	protected void checkIfTerminatable() {
		if(this.isTotalCommitDone()) {
			MsgBroker.reportCommitDone(this);
			this.state=STGState.COMMITTED;
			if(GlobalPara.sysConf.doMemoryManagement)
				this.terminateForMM();
			TPNode.LTM.STGPool.remove(this.key);
		}
	}
	
	public void reportCheckpointDone(SubTransaction t)
	{
		for(int i=0;i<trans.length;i++)
			if(t==trans[i]) {
				this.isCheckpointCompleted[i]=true;
				break;
			}
		checkIfTerminatable();
	}
	
	public int reportRestart()
	{
		this.state=STGState.ASKRESTART;
		MsgBroker.reportRestart(this);
		return 0;
	}
	
	
	/**
	 * 
	 * @param t
	 * @return true: if the canceled subtran is executed, false, if not 
	 */
	public boolean cancelCommit(SubTransaction t)
	{
		boolean result=false;
		if(t==null){
			GlobalPara.logger.warning("Receiving null subtransaction");
			return false;
		}
		for(int i=0;i<trans.length;i++)
			if(t==trans[i]){
				result=isCommitOK[i];
				isCommitOK[i]=false;
				if(this.trans[i].derivedTransactions!=null)
					this.trans[i].derivedTransactions.clear();
				break;
			}
		return result;
	}
	
	
	public boolean isTExeced(SubTransaction t)
	{
		for(int i=0;i<trans.length;i++)
			if(t==trans[i])
				return isCommitOK[i];
		System.out.println("[error]SubTransaction not find in this STG");
		return true;
	}
	
	
	private boolean isTotalCommitable()
	{
		boolean result=true;
		for(int i=0;i<isCommitOK.length;i++) {
			if(!isCommitOK[i]) {
				result=false;
				break;
			}
		}
		return result;
	}
	
	private boolean isTotalCommitDone()
	{
		boolean result=true;
		for(int i=0;i<this.isCommitCompleted.length;i++) {
			if(!isCommitCompleted[i]) {
				result=false;
				break;
			}
		}
		if(result){
			for(int i=0;i<this.isCheckpointCompleted.length;i++) {
				if(!isCheckpointCompleted[i]) {
					result=false;
					break;
				}
			}
		}
		
		return result;
	}
	
	
//	/**
//	 * [0] the first ST to be executed
//	 * [1] the second ST to be executed
//	 * ...
//	 * @param tg
//	 * @return
//	 */
//	private int[] getSequence()
//	{
//		SubTransaction[] tg= this.trans;
//		int[] stgHashCode=new int[tg.length];
//		TreeSet<Integer> sort=new TreeSet<Integer>();
//		for(int i=0;i<tg.length;i++)
//		{
//			stgHashCode[i]=LocalTransManager.getSTkey(tg[i]).hashCode();
//			sort.add(stgHashCode[i]);
//		}
//		Integer sortedResult[]=new Integer[tg.length];
//		sortedResult=sort.toArray(sortedResult);
//		int[] result=new int[tg.length];
//		for(int i=0;i<sortedResult.length;i++)
//		{
//			int tmp=sortedResult[i];
//			for(int j=0;j<stgHashCode.length;j++)
//			{
//				if(tmp==stgHashCode[j])
//					{
//					result[i]=j;
//					break;
//					}
//			}
//		}
//		return result;
//	}
	
	public String printDebugInfo()
	{
		String result= "STG with key=" + this.key + ", timestamp=" + this.timestamp + ", ";
		result+="GTPnodeID=" + this.GTPNodeID + ", ";
		result+="TPnodeID=" + this.TPnodeID + ", ";
		if(this.state!=null)
			result+="State=" + this.state.toString() + "\n";
		else
			result+="State=null \n";
		
		result+="Count of ST:" + this.trans.length + "\n";
		for(int i=0;i<this.trans.length;i++){
			String stResult=i+")ST isCommit="+this.isCommitOK[i]+", isReplicaCompleted="+this.isCommitCompleted[i]
			+", isCheckpointCompleted="+this.isCheckpointCompleted[i]+ ", ";
			if(this.trans[i].rootEntry==null)
				stResult+="rootDataEntry=null,";
			else {
				stResult+="rootDataEntry="+ this.trans[i].rootEntry.dataTable.tblName +":"+ this.trans[i].rootEntry.rowKey +",";
				stResult+="isdataloaded="+ this.trans[i].rootEntry.isDataLoaded+
					", isInTheQueue="+this.trans[i].rootEntry.isSTInOfList(this.trans[i])
					+", isHeadofQueue="+this.trans[i].rootEntry.isSTHeadOfList(this.trans[i])+"\n";
			}
			result+=stResult;
		}

		return result;
	}
	
	public String printInfo()
	{
		String result= this.key+ ":" + this.timestamp + "\n";
		result+="GTPnodeID:" + this.GTPNodeID + "\n";
		result+="TPnodeID:" + this.TPnodeID + "\n";
		result+="Count of ST:" + this.trans.length + "\n";
		if(this.state==null)
			return result;
		result+="State:" + this.state.toString() + "\n";
		result+="How many ST committed:";
		int count=0;
		String ctran="";
		for(int i=0;i<this.isCommitOK.length;i++)
		{
			if(isCommitOK[i])
				{
				ctran+=" " + i;
				count++;
				}
		}
		result+=" " + count + "\n";
		result+="The Committed STs are " + ctran + "\n";
		return result;
	}
}
